@@ -1,7 +1,7 @@
 /* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */
 
 /* Copyright (c) 2004 - 2006 Derek Foreman, Ben Jansens
-   Copyright (c) 2006 - 2014 Thomas Schmitt <scdbackup@gmx.net>
+   Copyright (c) 2006 - 2017 Thomas Schmitt <scdbackup@gmx.net>
    Provided under GPL version 2 or later.
 */
 
@@ -331,8 +331,31 @@ static int fifo_set_size(struct burn_source *source, off_t size)
 static void fifo_free(struct burn_source *source)
 {
 	struct burn_source_fifo *fs = source->data;
+	int wait_count;
+	static int wait_max = 30, wait_usleep = 100000;
 
 	burn_fifo_abort(fs, 0);
+	for (wait_count = 0; wait_count <= wait_max; wait_count++) {
+		if (fs->thread_is_valid <= 0)
+	break;
+		if (wait_count < wait_max)
+			usleep(wait_usleep);
+	}
+	if (wait_count > wait_max) {
+		/* The shoveler thread might still be active. If so, it would
+		   use invalid or inappropriate memory if the fifo would be
+		   disposed now. A memory and resource leak is the better
+		   option here.
+		*/
+		libdax_msgs_submit(libdax_messenger, -1,
+				   0x000201ab,
+				   LIBDAX_MSGS_SEV_WARNING,
+				   LIBDAX_MSGS_PRIO_HIGH,
+				   "Leaving burn_source_fifo object undisposed because it is possibly stuck but alive",
+				   0, 0);
+		return;
+	}
+
 	if (fs->inp != NULL)
 		burn_source_free(fs->inp);
 
@@ -356,13 +379,20 @@ int burn_fifo_source_shoveller(struct burn_source *source, int flag)
 	fs->thread_pid = getpid();
 	fs->thread_is_valid = 1;
 
+	/* Lock was obtained by async.c:add_worker() */
+	burn_async_manage_lock(BURN_ASYNC_LOCK_RELEASE);
+
 	bufsize = fs->chunksize * fs->chunks;
 	while (!fs->end_of_consumption) {
+		if (fs->do_abort)
+			goto emergency_exit;
 
 		/* wait for enough buffer space available */
 		wpos = fs->buf_writepos;
 		counted = 0;
 		while (1) {
+			if (fs->do_abort)
+				goto emergency_exit;
 			rpos = fs->buf_readpos;
 			diff = rpos - wpos;
 			trans_end = 0;
@@ -405,6 +435,8 @@ int burn_fifo_source_shoveller(struct burn_source *source, int flag)
 		}
 
 		/* Obtain next chunk */
+		if (fs->do_abort)
+			goto emergency_exit;
 		if (fs->inp->read != NULL)
 			ret = fs->inp->read(fs->inp,
 				 (unsigned char *) bufpt, fs->inp_read_size);
@@ -430,6 +462,8 @@ int burn_fifo_source_shoveller(struct burn_source *source, int flag)
 		fs->put_counter++;
 
 		/* activate read chunk */
+		if (fs->do_abort)
+			goto emergency_exit;
 		if (ret > fs->inp_read_size)
 					/* beware of ill custom burn_source */
 			ret = fs->inp_read_size;
@@ -463,8 +497,11 @@ int burn_fifo_source_shoveller(struct burn_source *source, int flag)
 		fs->end_of_input = 1;
 
 	/* wait for end of reading by consumer */;
-	while (fs->buf_readpos != fs->buf_writepos && !fs->end_of_consumption)
-			fifo_sleep(0);
+	while (fs->buf_readpos != fs->buf_writepos && !fs->end_of_consumption) {
+		if (fs->do_abort)
+			goto emergency_exit;
+		fifo_sleep(0);
+	}
 
 	/* destroy ring buffer */;
 	if (!fs->end_of_consumption)
@@ -481,8 +518,11 @@ int burn_fifo_source_shoveller(struct burn_source *source, int flag)
 			((size_t) fs->chunksize) * (size_t) fs->chunks, 0);
 	fs->buf = NULL;
 
+emergency_exit:;
+	burn_async_manage_lock(BURN_ASYNC_LOCK_OBTAIN);
 	fs->thread_handle= NULL;
 	fs->thread_is_valid = 0;
+	burn_async_manage_lock(BURN_ASYNC_LOCK_RELEASE);
 	return (fs->input_error == 0);
 }
 
@@ -524,6 +564,7 @@ struct burn_source *burn_fifo_source_new(struct burn_source *inp,
 	fs->thread_handle = NULL;
 	fs->thread_pid = 0;
 	fs->thread_is_valid = 0;
+	fs->do_abort = 0;
 	fs->inp = NULL; /* set later */
 	if (flag & 1)
 		fs->inp_read_size = 32 * 1024;
