@@ -51,7 +51,7 @@ sg_initialize()         performs global initialization of the SCSI transport
                         software components.
 
 sg_shutdown()           performs global finalizations and releases golbally
-                        aquired resources.
+                        acquired resources.
 
 sg_give_next_adr()      iterates over the set of potentially useful drive 
                         address strings.
@@ -60,7 +60,7 @@ scsi_enumerate_drives() brings all available, not-whitelist-banned, and
                         accessible drives into libburn's list of drives.
 
 sg_dispose_drive()      finalizes adapter specifics of struct burn_drive
-                        on destruction. Releases resources which were aquired
+                        on destruction. Releases resources which were acquired
                         underneath scsi_enumerate_drives().
 
 sg_drive_is_open()      tells wether libburn has the given drive in use.
@@ -201,7 +201,7 @@ static int linux_sg_auto_family = 1;
 
 /* Set this to 1 in order to accept any TYPE_* (see scsi/scsi.h) */
 /* But try with 0 first. There is hope via CDROM_DRIVE_STATUS. */
-/* !!! DO NOT SET TO 1 UNLESS YOU PROTECTED ALL INDISPENSIBLE DEVICES
+/* !!! DO NOT SET TO 1 UNLESS YOU PROTECTED ALL INDISPENSABLE DEVICES
        chmod -rw !!! */
 static int linux_sg_accept_any_type = 0;
 
@@ -1041,22 +1041,18 @@ static int is_scsi_drive(char *fname, int fd_in, int *bus_no, int *host_no,
 		{ret = 0; goto ex;}
 	}
 
-	if (sid_ret == -1 || sid.scsi_id < 0) {
-		/* ts A61211 : employ a more general ioctl */
-		/* ts B11001 : re-use fd */
-		ret = sg_obtain_scsi_adr_fd(fname, fd, bus_no, host_no,
-					   channel_no, target_no, lun_no);
-		if (ret>0) {
-			sid.host_no = *host_no;
-			sid.channel = *channel_no;
-			sid.scsi_id = *target_no;
-			sid.lun = *lun_no;
-		} else {
-			if (linux_sg_enumerate_debug)
-				fprintf(stderr,
-					"sg_obtain_scsi_adr_fd() failed\n");
-			{ret = 0; goto ex;}
-		}
+	/* ts A61211 : employ a more general ioctl */
+	/* ts B11001 : re-use fd */
+	/* ts B80902 : call unconditionally because host_no differs
+	               between SG_GET_SCSI_ID and SCSI_IOCTL_GET_IDLUN
+	*/
+	ret = sg_obtain_scsi_adr_fd(fname, fd, bus_no, host_no,
+				    channel_no, target_no, lun_no);
+	if (ret <= 0) {
+		if (linux_sg_enumerate_debug)
+			fprintf(stderr,
+				"sg_obtain_scsi_adr_fd() failed\n");
+		{ret = 0; goto ex;}
 	}
 
 	/* ts A60927 : trying to do locking with growisofs */
@@ -1076,16 +1072,6 @@ static int is_scsi_drive(char *fname, int fd_in, int *bus_no, int *host_no,
 		sg_release_siblings(sibling_fds, sibling_fnames,
 							&sibling_count);
 	}
-#ifdef SCSI_IOCTL_GET_BUS_NUMBER
-	if(*bus_no == -1)
-		*bus_no = 1000 * (sid.host_no + 1) + sid.channel;
-#else
-	*bus_no = sid.host_no;
-#endif
-	*host_no= sid.host_no;
-	*channel_no= sid.channel;
-	*target_no= sid.scsi_id;
-	*lun_no= sid.lun;
 	ret = 1;
 ex:;
 	if (fd_in < 0 && fd >= 0) {
@@ -1526,7 +1512,7 @@ int sg_initialize(char msg[1024], int flag)
 
 
 /** Performs global finalization of the SCSI transport adapter and eventually
-    needed operating system facilities. Releases globally aquired resources.
+    needed operating system facilities. Releases globally acquired resources.
     @param flag  unused yet, submit 0
     @return      1 = success, <=0 = failure
 */  
@@ -2231,6 +2217,10 @@ int sg_issue_command(struct burn_drive *d, struct command *c)
 		if (d->cancel)
 	break;
 		/* if ! done : loop for retry */;
+		if (!done && (burn_sg_log_scsi & 3)) {
+			scsi_log_text("+++ Repeating command", fp, 0);
+			scsi_log_cmd(c, fp, 0);
+		}
 	}
 
 	ret = 1;
@@ -2355,7 +2345,7 @@ ex:;
 
 /* ts B00115 */
 /* Return 1 if the given path leads to a regular file or a device that can be
-   seeked, read, and possibly written with 2 kB granularity. 
+   fseeked, read, and possibly written with 2 kB granularity. 
 */
 int burn_os_is_2k_seekrw(char *path, int flag)
 {
